#pragma once

#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/sprite.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/matrix.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/shader.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/shapes/triangle.hpp>

#include "sprite_data.hpp"


namespace BGTU::ComputerGraphicsLabWork::Impl {
	class Triangle : public RendererApi::Sprite<SpriteData> {
	public:
		const RendererApi::PointF<3> triangle_p0;
		const RendererApi::PointF<3> triangle_p1;
		const RendererApi::PointF<3> triangle_p2;

	private:
		const RendererApi::PointF<2> shader_p0;
		const RendererApi::PointF<2> shader_p1;
		const RendererApi::PointF<2> shader_p2;
		Utilities::Shader const *const shader;

	public:
		Triangle(
				RendererApi::PointF<3> triangle_p0,
				RendererApi::PointF<3> triangle_p1,
				RendererApi::PointF<3> triangle_p2,
				RendererApi::PointF<2> shader_p0,
				RendererApi::PointF<2> shader_p1,
				RendererApi::PointF<2> shader_p2,
				Utilities::Shader const *shader
		) : triangle_p0{triangle_p0}, triangle_p1{triangle_p1}, triangle_p2{triangle_p2},
			shader_p0{shader_p0}, shader_p1{shader_p1}, shader_p2{shader_p2}, shader{shader} {}

		void draw(BGTU::ComputerGraphicsLabWork::RendererApi::VirtualVoxelPainter *frame, SpriteData const *data) const override {
			auto ta = data->transform * this->triangle_p0;
			auto tb = data->transform * this->triangle_p1;
			auto tc = data->transform * this->triangle_p2;

			Utilities::Shapes::triangle_barycentric_interpolator<RendererApi::PointF<3>> interpolator{
					{ta.x, ta.y}, {this->shader_p0.x, this->shader_p0.y, ta.z},
					{tb.x, tb.y}, {this->shader_p1.x, this->shader_p1.y, tb.z},
					{tc.x, tc.y}, {this->shader_p2.x, this->shader_p2.y, tc.z},
			};

			Utilities::Shapes::iterate_triangle_fill(
					{static_cast<RendererApi::PointI<2>::component_t>(ta.x), static_cast<RendererApi::PointI<2>::component_t>(ta.y)},
					{static_cast<RendererApi::PointI<2>::component_t>(tb.x), static_cast<RendererApi::PointI<2>::component_t>(tb.y)},
					{static_cast<RendererApi::PointI<2>::component_t>(tc.x), static_cast<RendererApi::PointI<2>::component_t>(tc.y)},
					[&, this](bool is_edge, RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) {
						RendererApi::PointF<3> _tmp = interpolator.interpolate_point(x, y);
						frame->add_voxel(x, y, _tmp.z, this->shader->get_color(_tmp.x, _tmp.y));
					}
			);

		}
	};
}